// *****************************************************************************
// *                                                                           *
// *  BiqCrunch is a semidefinite-based solver for binary quadratic problems.  *
// *  It uses a branch-and-bound method featuring an improved semidefinite     *
// *  bounding procedure, mixed with a polyhedral approach. BiqCrunch uses     *
// *  particular input files format to describe the combinatorial problems.    *
// *                                                                           *
// *   Copyright (C) 2010-2014 Nathan Krislock, Jérôme Malick, Frédéric Roupin *
// *                                                                           *
// *                 http://www-lipn.univ-paris13.fr/BiqCrunch/                *
// *									       *
// *****************************************************************************
//									       *
//    This program is free software: you can redistribute it and/or modify     *
//    it under the terms of the GNU General Public License as published by     *
//    the Free Software Foundation, either version 3 of the License, or        *
//    (at your option) any later version.                                      *
//                                                                             *
//    This program is distributed in the hope that it will be useful,          *
//    but WITHOUT ANY WARRANTY; without even the implied warranty of           *
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
//    GNU General Public License for more details.                             *
//                                                                             *
//    You should have received a copy of the GNU General Public License        *
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.    *
//                                                                             *
// *****************************************************************************
// Authors : N. Krislock, F. Roupin 2012

#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <time.h>
#include <math.h>
#include <limits.h>

#include "bb.h"
#include "biqcrunch.h"

/**
 * Name of the heuristic printed in the output file
 */
char heur_name[] = "Max-Cut";


int *GWcut;

int BC_allocHeuristic(Problem *P0) {
    int N = P0->n - 1;
	int i;

	GWcut = (int *) malloc(sizeof(int) * (N + 1));
	for (i = 0; i < N + 1; i++)
		GWcut[i] = 0;

	return 1;
}

void BC_freeHeuristic() {
	free(GWcut);
}

double primalHeuristic(Problem *P0, int *x, double alpha);
double sdpBoundHeuristic(Problem *P0, BobNode *node, int *x, double alpha);
double rounding_heuristic(Problem *P0, Problem *P, BobNode *node, int *x);

/**
 * Generic call of the heuristics.
 * This function is called at different times:
 * - at the beginning of the algorithm (PRIMAL_HEUR)
 * - during the evaluation of a node bound (SDP_BOUND_HEUR)
 * - after the evaluation of a node (ROUNDING_HEUR)
 * It has to return the value obtained during the execution of the chosen
 * heuristic and the binary vector of the solution
 *
 * @param P0: the original problem in domain [-1:1]
 * @param P: the current problem in domain [-1:1] (NULL if heuristic_code equals to PRIMAL_HEUR)
 * @param node: the current node () (NULL if heuristic_code equals to PRIMAL_HEUR)
 * @param x: output binary vector of the solution obtained from the heuristic in domain [0:1]
 * @param heuristic_code: code that identify the timing of the call (PRIMAL_HEUR, SDP_BOUND_HEUR, ROUNDING_HEUR)
 * @return value of the solution found with the heuristic
 */
double BC_runHeuristic(Problem *P0, Problem *P, BobNode *node, int *x, int heuristic_code) {
    int N = P0->n - 1;
	double best = 0.0, heur_val = 0.0;
	int i;
	int temp_x[N]; // temporary vector to store heuristic solutions

	if (P0->max_problem) best = -BIG_NUMBER; else best = BIG_NUMBER;

	switch (heuristic_code) {
	case PRIMAL_HEUR:
		for (i = 0; i < 100; i++) {

			heur_val = primalHeuristic(P0, temp_x, ((double) i) / (double) 100);

            // copy the result only if it's better than the current and it's a
            // feasible solution
			if (((P0->max_problem && best < heur_val)
					|| (!P0->max_problem && best > heur_val))
					&& BC_isFeasibleSolution(temp_x)) {
				memcpy(x, temp_x, sizeof(int) * N);
				best = heur_val;
			}

		}
	
		break;
	case SDP_BOUND_HEUR:

		for (i = 0; i < 100; i++) {
			heur_val = sdpBoundHeuristic(P0, node, temp_x, ((double) i) / (double) 100); 

			if (((P0->max_problem && best < heur_val)
					|| (!P0->max_problem && best > heur_val))
					&& BC_isFeasibleSolution(temp_x)) {
				memcpy(x, temp_x, sizeof(int) * N);
				best = heur_val;
			}
		}

		break;
	case ROUNDING_HEUR:
		best = rounding_heuristic(P0, P, node, x);
		break;
	}

	return best;

}


double sdpBoundHeuristic(Problem *P0, BobNode * node, int * x, double alpha) {
	int i;

	int index = 0;

	for (i = 0; i < P0->n - 1; i++) {
		if (node->xfixed[i])
			x[i] = node->sol.X[i];
		else
			x[i] = 0;
	}

	for (i = 0; i < P0->n - 1; i++) {
		if (!node->xfixed[i]) {
			if (node->fracsol[i] > alpha)
				x[i] = 1;
			else
				x[i] = 0;
			index++;
		}

	}

	return BC_evaluateSolution(x);
}

double primalHeuristic(Problem *P0, int *x, double alpha) {
    int N = P0->n - 1;
	int i;
	double random[N];
	srand(time(NULL));

	for (i = 0; i < N; i++) {
		random[i] = ((double) rand() / (double) RAND_MAX);
	}

	for (i = 0; i < N; i++) {
		if (random[i] > alpha) {
			x[i] = 1;
		} else {
			x[i] = 0;
		}
	}

	return BC_evaluateSolution(x);
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void choleskyIncom(double *W, int n);
double GW(int *cut, double *Matrix, int N, double *Xnew);

double rounding_heuristic(Problem *P0, Problem *P, BobNode *node, int *X) {
	double val;
	int i;
	int index;

	val = GW(GWcut, P->Q, P->n, node->fracsol);
	index = 0;
	for (i = 0; i < P0->n - 1; i++) {
		if (!(node->xfixed[i])) {
			if (GWcut[P->n - 1] == 1) {
				X[i] = (1 + GWcut[index]) / 2;
			} else {
				X[i] = (1 - GWcut[index]) / 2;
			}
			index++;
		} else {
			X[i] = node->sol.X[i];
		}
	}

	val = BC_evaluateSolution(X);
	return val;
}

/************************ GW *********************************/
double GW(cut, Matrix, N, Xnew)
	int *cut;double *Matrix;int N;double *Xnew;
{

	// Incomplete Cholesky decomposition of Xnew, result is stored in Y

	int i, j, k, ic, jc;
	double entier;
	double entier1, entier2;
	double norm, inorm;
	double sca;
    double *Y = (double *) malloc(sizeof(double)*N*N);
	double e0[N];
	int e1[N], e2[N];

	// Initialize entier
	entier = 0.0;
	for (ic = 0; ic < N; ic++)
		for (jc = 0; jc < N; jc++)
			entier += Matrix[N * ic + jc] * cut[ic] * cut[jc];
//	printf("\nentier %lf\n", entier);
	// Y= Xnew;
	for (i = 0; i < N; i++)
		for (j = 0; j <= i; j++)
			Y[j * N + i] = Xnew[j * N + i];

	choleskyIncom(Y, N);

	entier2 = 0.0;
	// Goemans/Williamson primalization
	//

	srand(time(NULL));
	for (k = 0; k < 2000; k++) {
		norm = 0.0;
		////////////////////////////
		// use e0 for the First random hyperplane rounding (FR)
		if (k == 0) {
			for (i = 0; i < N; i++) {
				e0[i] = Y[i * N];
				norm += e0[i] * e0[i];
			}
		}
		///////////////////////////////////////////////
		else {
			// Compute the random hyperplane e0
			for (i = 0; i < N; i++) {
				e0[i] = 1.0 + (int) 100.0 * rand() / ((double) RAND_MAX + 1);
				norm += e0[i] * e0[i];
			}
		}
		inorm = 1.0 / sqrt(norm);
		for (i = 0; i < N; i++)
			e0[i] = inorm * e0[i];

		// Compute cut e1 generated by hyperplane e0
		for (i = 0; i < N; i++) {
			sca = 0.0;
			for (j = 0; j <= i; j++)
				sca += e0[j] * Y[j * N + i]; // Y is lower triangular
			if (sca < 0)
				e1[i] = -1;
			else
				e1[i] = 1;
		}

		// Compute value of cut e1
		entier1 = 0.0;
		for (i = 0; i < N; i++)
			for (j = 0; j < N; j++)
				entier1 += Matrix[N * i + j] * e1[i] * e1[j];

		// if value of new cut e1 is better than "cut"
		if (entier1 > entier2) {
			entier2 = entier1;
			for (j = 0; j < N; j++)
				e2[j] = e1[j];
		}
	}

	// Try to improve the best cut e2 that was found
	// by swapping each vertex to the other side of the partition

	int better = 1; // = >until no new improvment is made (FR)

	while (better) {
		better = 0;
		for (i = 0; i < N; i++) {
			for (j = 0; j < N; j++) {
				if (i == j)
					e1[j] = -e2[j];
				else
					e1[j] = e2[j];
			}

			//            entier1 = 0.0;
			entier1 = entier2;

			for (ic = 0; ic < N; ic++) {
				// Compute only the variation, since only e1[i]!=e2[i]    (FR)   O(n^3)=>O(n^2)

				entier1 -= Matrix[N * ic + i] * e2[ic] * e2[i];
				entier1 += Matrix[N * ic + i] * e1[ic] * e1[i];			

			}

			if (entier1 > entier2) {
				entier2 = entier1;
				for (jc = 0; jc < N; jc++)
					e2[jc] = e1[jc];
				better = 1;
			}

		}
	}

	// If e2 is better than "cut"
	if (entier2 > entier) {
		entier = entier2;
		for (jc = 0; jc < N; jc++)
			cut[jc] = (int) e2[jc];
//		printf("GW cut improved:  %.0f\n", entier);

	}

    free(Y);

	return entier;

}

/************************ choleskyIncom *********************************/
void choleskyIncom(double *W, int n) {
	// Construit un champ de vecteur V ayant pour matrice de Gram W n*n
	int i, j, k;
	double g, h;

	for (j = 0; j < n; j++) {
		// Pour chaque ligne de la matrice
		g = W[n * j + j];
		for (k = 0; k < j; k++)
			g -= W[n * k + j] * W[n * k + j];
		if (g <= 0.0) {
			for (i = j; i < n; i++)
				W[n * j + i] = 0.0;
		} else {
			W[n * j + j] = g = sqrt(fabs(g));
			for (i = j + 1; i < n; i++) {
				h = W[n * j + i];
				for (k = 0; k < j; k++)
					h -= W[n * k + i] * W[n * k + j];
				W[n * j + i] = h / g;
			}
		}
	}
	for (j = 0; j < n; j++)
		for (i = 0; i < j; i++)
			W[n * j + i] = 0.0;
}

