// *****************************************************************************
// *                                                                           *
// *  BiqCrunch is a semidefinite-based solver for binary quadratic problems.  *
// *  It uses a branch-and-bound method featuring an improved semidefinite     *
// *  bounding procedure, mixed with a polyhedral approach. BiqCrunch uses     *
// *  particular input files format to describe the combinatorial problems.    *
// *                                                                           *
// *   Copyright (C) 2010-2014 Nathan Krislock, Jérôme Malick, Frédéric Roupin *
// *                                                                           *
// *                 http://www-lipn.univ-paris13.fr/BiqCrunch/                *
// *									       *
// *****************************************************************************
//									       *
//    This program is free software: you can redistribute it and/or modify     *
//    it under the terms of the GNU General Public License as published by     *
//    the Free Software Foundation, either version 3 of the License, or        *
//    (at your option) any later version.                                      *
//                                                                             *
//    This program is distributed in the hope that it will be useful,          *
//    but WITHOUT ANY WARRANTY; without even the implied warranty of           *
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
//    GNU General Public License for more details.                             *
//                                                                             *
//    You should have received a copy of the GNU General Public License        *
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.    *
//                                                                             *
// *****************************************************************************

#include <stdio.h>
#include <math.h>
#include <sys/resource.h>

#include "bb.h"
#include "biqcrunch.h"

extern FILE *output;
extern BiqCrunchParameters params;
extern Problem *SP;
extern double *SubMat; // Construction matrices (size BobSizePb)
extern double *SubMatLin; // Construction vector (size BobSizePb)

//=> F.R. 2014 Sparse
extern Sparse SubMatS; // Construction matrice, sparse representation (maxsize BobSizePb)

extern double *tempMatrix; // Construction vector (size BobSizePb + 1)
extern int evalcount; // Number of node evaluations
extern double TEMPI; // initial time
extern int stopped;
extern double best_eval;
extern int current_bound;


/*
 * Bob user function which initialize the problem, allocate the structures 
 * and evaluate the root node
 * @param n: number of arguments on the command line
 * @param v: array of the command line arguments
 */
void Bob_Init(int n, char **v) {
	BobSolution bs;
	int i;

    TEMPI = temps_CPU(); // start the timer

	// Control the command line arguments
	if (n != 2) {
		fprintf(stderr,
				"Usage : biqcrunch [-v (0|1)] file.bc file.param\n");
		exit(1);
	}
	// Create the output file
	if (!createOutputFile(v[0])) {
		fprintf(stderr, 
                "Can't create output file in the current folder\n");
		exit(1);
	}

	// Read the input file instance
	Read_Data(v[0]);

	// Read the parameters from a user file
    readParameter(v[1]);

	// Initialize the Bob first solution (no variables set to 1)
	for (i = 0; i < BobPbSize; i++)
		bs.X[i] = 0;

	// Update Bob solution with the initial solution
    Bob_ULBInit(BIG_NUMBER, &bs);

	// Allocate the global matrices and the vector used to compute the submatrix 
    // in the evaluate function
	alloc_matrix(tempMatrix, SP->n, double);
	alloc_matrix(SubMat, BobPbSize, double);
	alloc_vector(SubMatLin, BobPbSize, double);

//=> F.R. Sparse
	alloc_vector(SubMatS.i, (BobPbSize * BobPbSize) / 2 + BobPbSize, int);
	alloc_vector(SubMatS.j, (BobPbSize * BobPbSize) / 2 + BobPbSize, int);
	alloc_vector(SubMatS.val, (BobPbSize * BobPbSize) / 2 + BobPbSize, double);

	// Allocate the global matrices and vectors used to compute the bound
	allocSDPbound(SP);

	// Allocate the heuristic global structures
	if (!BC_allocHeuristic(SP)) {
		fprintf(stderr, "\nError: cannot allocate heuristic data structure\n");
		exit(1);
	}

	// Redirect the Fortran output to /dev/null to have a cleaner output in the 
        // command line
	redirect_output_();

	// Initialize the problem
	Init_UQP();
}


//=========Initialization : root node ==================//
void Init_UQP() {
	int i;
	double eval;
	double heur_val;
	int x[BobPbSize];

	BobRoot = Bob_NodeAlloc(0);

	if (BobRoot == NULL) {
		printf("[0]:Error not enough memory\n");
		exit(1);
	}

	for (i = 0; i < BobPbSize; i++) {
		BobRoot->xfixed[i] = 0;
		BobRoot->sol.X[i] = 0;
		x[i] = 0;
	}

	BobRoot->level = 0;
	heur_val = BIG_NUMBER; // Current solution may not be feasible

    // run the heuristic to get initial solution
	if (params.heur_1)
		heur_val = BC_runHeuristic(SP, NULL, BobRoot, x, PRIMAL_HEUR);

	addSolution(BobRoot, -heur_val, x, NEWSOL_HEUR1);
	evalcount = 0;
	eval = Evaluate(BobRoot);
	best_eval = eval;
	if (!SP->max_problem) {
		Bob_PRICREAT(BobRoot->Pri, ceil(eval), 0, 0);
	} else 
        Bob_PRICREAT(BobRoot->Pri, (int) eval, 0, 0);
}



/*
 * Bob user defined function that generate new children.
 * The function generate two new children of the current node of the 
 * branch-and-bound tree.
 * The branching is made on the variable that has the fractional value closest to 0.5
 * @param Nodp: current node of the branch-and-bound tree
 * @param Depth: depth of the node in the branch-and-bound tree
 * @param ExpCt:
 */
void Bob_GenChild(BobNode *Nodp, int Depth, BobTExpCt *ExpCt) {
	BobNode *node;
	double eval;
	int i;
	int j;

	// If the algorithm stops before finding the optimal solution, search in the 
    // nodes queue for the best bound (if the algorithm stops at the root node the 
    // queue is empty and the best bound is the root node bound)
	if (stopped || params.root
			|| (params.time_limit > 0
					&& (temps_CPU() - TEMPI) > params.time_limit)) {
		int nb = Bob_PRIEVAL(Nodp->Pri);
		stopped = 1;
		if ( current_bound > nb)
		 	{
				best_eval = nb;
				current_bound = nb;
			}
		Bob_NodeFree(Nodp);
		return;
	}

// If it's a leaf of the branch-and-bound tree add the solution found 
// and don't branch
	if ((Nodp->level) == BobPbSize) {
		addSolution(Nodp, (double) Bob_PRIEVAL(Nodp->Pri), Nodp->sol.X, NEWSOL_LEAF);
		Bob_NodeFree(Nodp);
		return;
	}

// Otherwise we have to branch
// We search for the most fractional variable
	int ic = 0;  // x[ic] is the variable to split on
	double minValue = 2.0;

	double mean = 0.5;
/*
// Average value
	int counter = 0;
	double mean = 0.0;
	for (i = 0; i < BobPbSize; i++) {
		if (!(Nodp->xfixed[i])) {
			mean += Nodp->fracsol[i];
			counter++;
		}
	}

	mean /= counter;
*/

	for (i = 0; i < BobPbSize; i++) {
//		fprintf(output, "\nfracsol[%d] = %.2f\n", i, Nodp->fracsol[i]);
		if (!(Nodp->xfixed[i]) && fabs(mean - Nodp->fracsol[i]) < minValue) {
			ic = i;
			minValue = fabs(mean - Nodp->fracsol[ic]);
		}
	}

	fprintf(output, "\nfracsol[%d] = %.2f\n", ic, Nodp->fracsol[ic]);

	// add two nodes to the search tree
	for (j = 0; j <= 1; j++) {

        // allocate memory for the new node
		node = (BobNode *) Bob_NodeAlloc(0);
		node->level = Nodp->level + 1;

        // copy the solution information from the parent Nodp
		for (i = 0; i < BobPbSize; i++) {
			node->xfixed[i] = Nodp->xfixed[i];
			if (node->xfixed[i])
				node->sol.X[i] = Nodp->sol.X[i];
			else
				node->sol.X[i] = 0;
		}

        // split on node ic
		node->xfixed[ic] = 1;
		node->sol.X[ic] = j;
		fprintf(output, "fixed : X[%d] =  %d \n", ic, j);

		eval = Evaluate(node);

		int evali;
		if (SP->max_problem) evali = (int) eval; else evali = (int) ceil(eval);		

		Bob_PRICREAT(node->Pri, evali, 0, Depth + 1);

		fprintf(output, "Depth = %d\n", node->level);
		if (Bob_ULBSup(Bob_PRIEVAL(node->Pri))) {
			/* Branching */
			if (SP->max_problem)
				fprintf(output, "Branching!\nBound = %d Beta = %d \n",
						-Bob_PRIEVAL(node->Pri), -Bob_ULBGet());
			else
				fprintf(output, "Branching!\nBound = %d Beta = %d \n",
						Bob_PRIEVAL(node->Pri), Bob_ULBGet());
			if (Bob_ExpCtrl(Depth + 1, ExpCt))
				Bob_GenChild(node, Depth + 1, ExpCt);
			else
				Bob_GPQIns(node);

		} else {
			if (SP->max_problem)
				fprintf(output, "Bound = %d Beta = %d \n", -Bob_PRIEVAL(node->Pri),
						-Bob_ULBGet());
			else
				fprintf(output, "Bound = %d Beta = %d \n", Bob_PRIEVAL(node->Pri),
						Bob_ULBGet());
			/* prune */
			Bob_NodeFree(node);
		}
	}

	Bob_NodeFree(Nodp);

}

/*
 * Bob function called at the end of the execution.
 * This function print the statistics of the execution, such as the number of 
 * nodes evaluated, the best solution found, the nodes best bound, the CPU time.
 * The Bob_end function is also used to free the memory allocated by the program.
 */
void Bob_End() {
	int i;
	// Best solution found
	int best_sol = (SP->max_problem) ? -Bob_ULBGet() : Bob_ULBGet();
	// Best bound in the queue of the tree nodes
	double best_bound = (SP->max_problem) ? -best_eval : best_eval;
	// The gap between the best solution found and the bst bound
	double gap = (
			best_sol != 0 ?
					(best_bound - (double) best_sol) / (double) best_sol : 1.0)
			* 100.0;

	// Print the results on the standard output
	printf("\bNodes = %d; Root node bound = %.2lf\n", evalcount, best_bound);
	if (!stopped) {
        if (Bob_ULBGet() == BIG_NUMBER)
            printf("Problem is infeasible.\n");
        else { 
            printf("%s value = %d\n", 
                    (SP->max_problem) ? "Maximum" : "Minimum", best_sol);
            printf("Solution = { ");
            for (i = 0; i < BobPbSize; i++)
                if (BobSol->X[i] == 1)
                    printf("%d ", i + 1);
            printf("}\n");
        }
	} else {
		if (params.root)
			printf("Best value = %d\nRoot node bound = %.2lf\nGap = %.2lf%%\n",
					best_sol, best_bound, gap);
		else
			printf("Best value = %d\nCurrent bound = %.0lf\nGap = %.2lf%%\n",
					best_sol, best_bound, gap);
	}
	printf("CPU time = %.4f s\n", temps_CPU() - TEMPI);

	// Print the results in the output file
	fprintf(output,
            "\nNodes = %d\n", evalcount);
	if (!stopped) {
        if (Bob_ULBGet() == BIG_NUMBER)
            fprintf(output,
                    "Problem is infeasible.\n");
        else { 
            fprintf(output,
                    "%s value = %d\nRoot node bound = %.2lf\nSolution = { ", 
                    (SP->max_problem) ? "Maximum" : "Minimum", best_sol, best_bound);
            for (i = 0; i < BobPbSize; i++)
                if (BobSol->X[i] == 1)
                    fprintf(output,
                            "%d ", i + 1);
            fprintf(output,
                    "}\n");
        }
	} else {
		if (params.root)
			fprintf(output,
					"Best value = %d\nRoot node bound = %.2lf\nGap = %.2lf%%\n",
					best_sol, best_bound, gap);
		else
			fprintf(output,
					"Best value = %d\nCurrent bound = %.0lf\nGap = %.2lf%%\n",
					best_sol, best_bound, gap);
	}
	fprintf(output, "CPU time = %.4f s\n", temps_CPU() - TEMPI);

	// Free the memory
	freeMemory();

	// Close the output file
	closeOutputFile();
}




double temps_CPU() 
{
	struct rusage usage;
	getrusage(RUSAGE_SELF, &usage);
	return usage.ru_utime.tv_sec + (usage.ru_utime.tv_usec * 1e-6);
}
