// *****************************************************************************
// *                                                                           *
// *  BiqCrunch is a semidefinite-based solver for binary quadratic problems.  *
// *  It uses a branch-and-bound method featuring an improved semidefinite     *
// *  bounding procedure, mixed with a polyhedral approach. BiqCrunch uses     *
// *  particular input files format to describe the combinatorial problems.    *
// *                                                                           *
// *   Copyright (C) 2010-2014 Nathan Krislock, Jérôme Malick, Frédéric Roupin *
// *                                                                           *
// *                 http://www-lipn.univ-paris13.fr/BiqCrunch/                *
// *									       *
// *****************************************************************************
//									       *
//    This program is free software: you can redistribute it and/or modify     *
//    it under the terms of the GNU General Public License as published by     *
//    the Free Software Foundation, either version 3 of the License, or        *
//    (at your option) any later version.                                      *
//                                                                             *
//    This program is distributed in the hope that it will be useful,          *
//    but WITHOUT ANY WARRANTY; without even the implied warranty of           *
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
//    GNU General Public License for more details.                             *
//                                                                             *
//    You should have received a copy of the GNU General Public License        *
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.    *
//                                                                             *
// *****************************************************************************

#include <math.h>

#include "bb.h"
#include "biqcrunch.h"

extern Inequality *Cuts; // vector of triangle inequality constraints
extern double *PX; // projected matrix computed by projSDP
extern double *X; // Matrix to work on : modified by proj
extern double *W; // contains the eigenvalues
extern double *Z; // contains the eigenvectors
extern int *ISUPPZ; // dim = 2*max(1,M), support of the eigvecs. in Z
extern double *WORK; // dim = LWORK
extern int *IWORK; // dim = LIWORK
extern int sizeWORK, sizeIWORK;
extern double scaleIneq; // scales the inequality constraints (and dual vars)
extern double scaleEq; // scales the equality constraints (and dual vars)


// computes gradient and function values for x
/*************** Simulator (evaluates function and gradient) ******************/
void sim(Problem *P, int NumIneq, double *yy, double *ff, double *gg, double alpha) 
{
	double alphainv = 1.0 / alpha;

	int N = P->n;
	double *Matrix = P->Q;

	int N2 = N * N;
	int incx = 1;
	int incy = 1;
	dcopy_(&N2, Matrix, &incx, X, &incy); // X = Matrix

	// X = X - scaleEq * B^*(yy)
	B(TRANSP, P, yy, ff, gg, alpha);
	A(TRANSP, P, NumIneq, yy, ff, gg, alpha);

	projSDP(ff, X,  PX, N); // returns *ff = 0.5*|X_+|_F^2

	*ff = *ff * 1000.0 * (alphainv / 1000.0); // now *ff = 0.5*|X_+|_F^2/alpha

	B(NOTRANSP, P, yy, ff, gg, alpha);
	A(NOTRANSP, P, NumIneq, yy, ff, gg, alpha);

	*ff += alpha * (0.5 * N * N);
}


/***********************  projSDP  ************************/
void projSDP(double *ff, double *A, double *Xplus, int N) 
{
	int i, j, k;
	double bornesup = 0.0;
	double borneinf = 1e-8;
	double lign;

    // Operator norm for A to get a upper bound of the spectral radius of A
	for (i = 0; i < N; i++) {
		lign = 0.0;
		for (j = 0; j < i; j++)
			lign += fabs(A[i + N * j]);
		for (j = i + 1; j < N; j++)
			lign += fabs(A[j + N * i]);
		lign += fabs(A[i + N * i]);
		if (lign > bornesup)
			bornesup = lign;
	}
	if (bornesup < borneinf)
		bornesup = 2.0 * borneinf;

    char JOBZ = 'V';
    char RANGE = 'V';
    char UPLO = 'L';
    int LDA = N;
    double VL = borneinf;
    double VU = bornesup;
    int IL = 0;
    int IU = 0;
    double ABSTOL = 1e-8;
	int M = 0;
    // double *W
    // double *Z
    int LDZ = N;
    // int ISUPPZ;
    // double *WORK;
    int LWORK = sizeWORK;
    // int *IWORK;
    int LIWORK = sizeIWORK;
    int INFO;

    // gets the eigenvalues (positive ones only) and eigenvectors associated
	dsyevr_(&JOBZ, &RANGE, &UPLO, &N, A, &LDA, &VL, &VU, &IL, &IU, &ABSTOL, &M,
			W, Z, &LDZ, ISUPPZ, WORK, &LWORK, IWORK, &LIWORK, &INFO);

    // the M columns of Z contain the orthonormal eigenvectors of the 
    // matrix A corresponding to the positive eigenvalues, the i-th
    // column of Z holding the eigenvector associated with W(i).

	*ff = 0.0;
	for (i = 0; i < N; i++) {

		//compute ff = ||Xplus||^2
		if (i < M && W[i] > borneinf)
			*ff += W[i] * W[i];

		//computes Xplus (terms of the lower triangular matrix)
		for (j = 0; j <= i; j++) {
			Xplus[i + j * N] = 0.0;
			for (k = 0; k < M; k++) {
				if (W[k] > borneinf) {
					Xplus[i + j * N] += Z[i + k * N] * Z[j + k * N] * W[k];
				}
			}
		}
	}

	// ff = 0.5*||Xplus||^2
	*ff *= 0.5;

}




/***************** A *********************/
void A(int mode, Problem *P, int NumIneq, double *yy, double *ff, double *gg,
		double alpha)
// Constraint:  A(X) <= a
//
// Evaluates:
//      if mode == TRANSP:
//          X = X + scaleIneq * A^*(yy)
//
//      if mode == NOTRANSP:
//          ff = ff + scaleIneq * <a,yy>
//          gg = scaleIneq*( a + A(PX/alpha) )
//
{
	double temp, alphainv = 1.0 / alpha, dd;
	int type, ii, jj, kk, ineq;

	int N = P->n, entry;

	if (mode == TRANSP) {
		for (ineq = 0; ineq < P->mA; ineq++) // for each equality constraint
        {
			temp = scaleEq * yy[P->mB + ineq];

			// Using sparse representation
			for (entry = 0; entry < P->As[ineq].nnz; entry++) // for each nonzero entry
            {
				ii = P->As[ineq].i[entry];
				jj = P->As[ineq].j[entry];
				dd = P->As[ineq].val[entry];
				if (ii == jj)
					X[ii + jj * N] -= temp * dd;
				else {
					X[ii + jj * N] -= temp * dd;
					X[jj + ii * N] -= temp * dd;
				}
			}
		}

		for (ineq = 0; ineq < NumIneq; ineq++) {

			type = Cuts[ineq].type;
			ii   = Cuts[ineq].i;
			jj   = Cuts[ineq].j;
			kk   = Cuts[ineq].k;

			temp = 0.5 * scaleIneq * yy[P->mB + P->mA + ineq];

			switch (type) {
			case 1:
				X[ii + jj * N] += temp;
				X[jj + ii * N] += temp;
				X[ii + kk * N] += temp;
				X[kk + ii * N] += temp;
				X[jj + kk * N] += temp;
				X[kk + jj * N] += temp;
				break;
			case 2:
				X[ii + jj * N] += temp;
				X[jj + ii * N] += temp;
				X[ii + kk * N] -= temp;
				X[kk + ii * N] -= temp;
				X[jj + kk * N] -= temp;
				X[kk + jj * N] -= temp;
				break;
			case 3:
				X[ii + jj * N] -= temp;
				X[jj + ii * N] -= temp;
				X[ii + kk * N] += temp;
				X[kk + ii * N] += temp;
				X[jj + kk * N] -= temp;
				X[kk + jj * N] -= temp;
				break;
			default: //type == 4
				X[ii + jj * N] -= temp;
				X[jj + ii * N] -= temp;
				X[ii + kk * N] -= temp;
				X[kk + ii * N] -= temp;
				X[jj + kk * N] += temp;
				X[kk + jj * N] += temp;
			}
		}

	} else { // mode==NOTRANSP

		for (ineq = 0; ineq < P->mA; ineq++) // for each equality constraint
        {
			*ff += scaleEq * P->a[ineq] * yy[P->mB + ineq];

			gg[P->mB + ineq] = 0.0;
			for (entry = 0; entry < P->As[ineq].nnz; entry++) // for each nonzero entry
					{
				ii = P->As[ineq].i[entry];
				jj = P->As[ineq].j[entry];
				dd = P->As[ineq].val[entry];
				if (ii == jj)
					gg[P->mB + ineq] -= dd * PX[ii + jj * N];
				else
					gg[P->mB + ineq] -= 2.0 * dd * PX[ii + jj * N];
			}
			gg[P->mB + ineq] *= alphainv;
			gg[P->mB + ineq] += P->a[ineq];
			gg[P->mB + ineq] *= scaleEq;
		}

		for (ineq = 0; ineq < NumIneq; ineq++) {

			type = Cuts[ineq].type;
			ii   = Cuts[ineq].i;
			jj   = Cuts[ineq].j;
			kk   = Cuts[ineq].k;

			*ff += scaleIneq * yy[P->mB + P->mA + ineq];

			switch (type) {
			case 1:
				temp = PX[ii + jj * N] + PX[ii + kk * N] + PX[jj + kk * N];
				break;
			case 2:
				temp = PX[ii + jj * N] - PX[ii + kk * N] - PX[jj + kk * N];
				break;
			case 3:
				temp = -PX[ii + jj * N] + PX[ii + kk * N] - PX[jj + kk * N];
				break;
			default: // type == 4
				temp = -PX[ii + jj * N] - PX[ii + kk * N] + PX[jj + kk * N];
			}
			gg[P->mB + P->mA + ineq] = (temp * alphainv + 1.0) * scaleIneq;
		}

	} // end if
}




/***************** B *********************/
void B(int mode, Problem *P, double *yy, double *ff, double *gg, double alpha)
// Constraint:  B(X) = b
//
// Evaluates:
//      if mode == TRANSP:
//          X = X - scaleEq * B^*(yy)
//
//      if mode == NOTRANSP:
//          ff = ff + scaleEq * <b,yy>
//          gg = scaleEq*( b - B(PX/alpha) )
//
{
	double alphainv = 1.0 / alpha;
	int ii, jj, eq, entry;
	double temp, dd;
	//int incx = 1;
	//int incy = 1;

	int N = P->n;

	if (mode == TRANSP) {
		for (eq = 0; eq < P->mB; eq++){ // for each equality constraint
			temp = scaleEq * yy[eq];

			// Using sparse representation
			for (entry = 0; entry < P->Bs[eq].nnz; entry++) // for each nonzero entry
					{
				ii = P->Bs[eq].i[entry];
				jj = P->Bs[eq].j[entry];
				dd = P->Bs[eq].val[entry];
				if (ii == jj)
					X[ii + jj * N] -= temp * dd;
				else {
					X[ii + jj * N] -= temp * dd;
					X[jj + ii * N] -= temp * dd;
				}
			}
		}

	} else { // mode==NOTRANSP

		for (eq = 0; eq < P->mB; eq++) {// for each equality constraint

			*ff += scaleEq * P->b[eq] * yy[eq];

			gg[eq] = 0.0;
			for (entry = 0; entry < P->Bs[eq].nnz; entry++) // for each nonzero entry
					{
				ii = P->Bs[eq].i[entry];
				jj = P->Bs[eq].j[entry];
				dd = P->Bs[eq].val[entry];
				if (ii == jj)
					gg[eq] -= dd * PX[ii + jj * N];
				else
					gg[eq] -= 2.0 * dd * PX[ii + jj * N];
			}
			gg[eq] *= alphainv;
			gg[eq] += P->b[eq];
			gg[eq] *= scaleEq;
		}

	} // end if
}

