// *****************************************************************************
// *                                                                           *
// *  BiqCrunch is a semidefinite-based solver for binary quadratic problems.  *
// *  It uses a branch-and-bound method featuring an improved semidefinite     *
// *  bounding procedure, mixed with a polyhedral approach. BiqCrunch uses     *
// *  particular input files format to describe the combinatorial problems.    *
// *                                                                           *
// *   Copyright (C) 2010-2014 Nathan Krislock, Jérôme Malick, Frédéric Roupin *
// *                                                                           *
// *                 http://www-lipn.univ-paris13.fr/BiqCrunch/                *
// *									       *
// *****************************************************************************
//									       *
//    This program is free software: you can redistribute it and/or modify     *
//    it under the terms of the GNU General Public License as published by     *
//    the Free Software Foundation, either version 3 of the License, or        *
//    (at your option) any later version.                                      *
//                                                                             *
//    This program is distributed in the hope that it will be useful,          *
//    but WITHOUT ANY WARRANTY; without even the implied warranty of           *
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
//    GNU General Public License for more details.                             *
//                                                                             *
//    You should have received a copy of the GNU General Public License        *
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.    *
//                                                                             *
// *****************************************************************************

#include <math.h>
#include <string.h>

#include "bb.h"
#include "biqcrunch.h"


extern BiqCrunchParameters params;
extern double *Xnew; // Stores new X
extern int count; // major iteration counter
extern double gradInorm; // norm of the gradient of f for equality constraints
extern Inequality *Cuts; // vector of triangle inequality constraints
extern Inequality *List; // list of inequalities
extern double *Xold; // Stores previous X
extern double *Xtest; // Stores inequality testing X
extern double minAllIneq; // value of the minimum inequality : feasible when minAllIneq >= 0
extern int NIneq; // current number of inequalities used in optimization
extern int NumSubtracted;
extern int NumAdded;



/******************** Add and Remove Inequalities ***************************/
void AddSubInequalities(Problem *P, double *y, int N, double seuil) 
{
	int ii, jj, kk, type;
	double ineqvalue; // value of the inequality:  ineqvalue >= 0 if feasible
	double testineqvalue;
	int ineq; // index for inequalities
	int yindex; // index for dual multipliers
	int ListCount, ListSize;
	double LeastViolatedIneqValue = -INFINITY;
	int LeastViolatedIneq = 0;

	// Initialize Xold as Xnew in the first iteration
	if (count == 1)
		memccpy(Xold, Xnew, N * N, sizeof(double));

	// Computing Xtest and store Xold
	double lambda = 0.2; // use lambda fraction of Xold
	for (ii = 0; ii < N; ii++) {
		for (jj = 0; jj <= ii; jj++) {
			Xtest[ii+jj*N] = lambda * Xold[ii+jj*N] 
                + (1.0-lambda) * Xnew[ii+jj*N];
			Xold[ii+jj*N] = Xnew[ii+jj*N]; // Xold stores only last iterate
		}
	}

    // Loop through all inequalities
    yindex = P->mB + P->mA;
    ListSize = 0;
	minAllIneq = 1e100;
	for (type = 1; type <= 4; type++) {
    for (ii = 0; ii < N;  ii++) {
    for (jj = 0; jj < ii; jj++) {
    for (kk = 0; kk < jj; kk++) {

        ineqvalue     = eval_ineq(Xnew,  N, type, ii, jj, kk);
        testineqvalue = eval_ineq(Xtest, N, type, ii, jj, kk);

        // keep track of the minimum value of ineqvalue
		minAllIneq = (ineqvalue < minAllIneq) ? ineqvalue : minAllIneq;

        if (testineqvalue < seuil)
        {
			// (1) put first NumToAdd violated inequalities in list and 
            //     keep track of the least violated inequality
			if (ListSize < params.cuts) {

				// add ineq to the end of List
                List[ListSize].type = type;
                List[ListSize].i    = ii;
                List[ListSize].j    = jj;
                List[ListSize].k    = kk;
                List[ListSize].value = testineqvalue;

				// update LeastViolatedIneq
				if (ListSize == 0 || testineqvalue > LeastViolatedIneqValue) {
					LeastViolatedIneqValue = testineqvalue;
					LeastViolatedIneq = ListSize;
				}

				ListSize++;
			} 
            else if (testineqvalue < LeastViolatedIneqValue) 
            {
				// (2) if you find an inequality is violated more than the 
                //     least violated inequality, then add it to the list, 
                //     remove the least violated inequality, and find 
                //     the new least violated inequality in the list

				// add ineq to the list, replacing LeastViolatedIneq
				List[LeastViolatedIneq].type = type;
				List[LeastViolatedIneq].i    = ii;
				List[LeastViolatedIneq].j    = jj;
				List[LeastViolatedIneq].k    = kk;
                List[LeastViolatedIneq].value = testineqvalue;

				// update LeastViolatedIneq
				LeastViolatedIneqValue = -100.0;
				for (ListCount = 0; ListCount < ListSize; ListCount++) 
                {
                    // if ineq is less violated
					if (List[ListCount].value > LeastViolatedIneqValue) 
                    {
						LeastViolatedIneqValue = List[ListCount].value;
						LeastViolatedIneq = ListCount;
                    }
                }
            }
        }
    } // kk loop
    } // jj loop
    } // ii loop
    } // type loop


    // Decide which cuts that were previously added to remove
    NumSubtracted = 0;
	yindex = P->mB + P->mA;
    int next_ineq = 0;
    for (ineq = 0; ineq < NIneq; ineq++) {
        type = Cuts[ineq].type;
        ii   = Cuts[ineq].i;
        jj   = Cuts[ineq].j;
        kk   = Cuts[ineq].k;

        ineqvalue = eval_ineq(Xnew, N, type, ii, jj, kk);

        // store the dual multiplier
        Cuts[ineq].y = y[yindex];  yindex++;

        // remove inequality if multiplier is small 
        // and inequality is large
        if (Cuts[ineq].y < 1e-8 && ineqvalue > gradInorm / 10.0) 
        {
            NumSubtracted++;
        } 
        else // keep inequality
        {
            Cuts[next_ineq].type = Cuts[ineq].type;
            Cuts[next_ineq].i    = Cuts[ineq].i;
            Cuts[next_ineq].j    = Cuts[ineq].j;
            Cuts[next_ineq].k    = Cuts[ineq].k;
            Cuts[next_ineq].y    = Cuts[ineq].y;
            next_ineq++;
        }
    }

    // Add List to Cuts
    NumAdded = 0;
	for (ListCount = 0; ListCount < ListSize; ListCount++) {
        // Stop if we have reached the maximum number of cuts
        // we can add
        if (next_ineq == MaxNineqAdded) break;
        // Check if inequality is already included in Cuts
        int found_ineq = 0;
        for (ineq = 0; ineq < NIneq-NumSubtracted; ineq++) {
            if (Cuts[ineq].type == List[ListCount].type &&
                Cuts[ineq].i == List[ListCount].i &&
                Cuts[ineq].j == List[ListCount].j &&
                Cuts[ineq].k == List[ListCount].k) 
            {
                found_ineq = 1;
            }
        }
        // If inequality not already in Cuts, add it to Cuts
        if (!found_ineq)
        {
            Cuts[next_ineq].type = List[ListCount].type;
            Cuts[next_ineq].i    = List[ListCount].i;
            Cuts[next_ineq].j    = List[ListCount].j;
            Cuts[next_ineq].k    = List[ListCount].k;
            Cuts[next_ineq].y    = 0.0;
            next_ineq++;
            NumAdded++;
        }
	}

}


/************************ eval_ineq *********************************/
double eval_ineq(double *XX, int N, int type, int ii, int jj, int kk) 
{
    double ineqvalue;

    // compute the inequality
    switch (type) {
    case 1:
			ineqvalue =  XX[ii+jj*N] + XX[ii+kk*N] + XX[jj+kk*N] + 1.0;
			break;
		case 2:
			ineqvalue =  XX[ii+jj*N] - XX[ii+kk*N] - XX[jj+kk*N] + 1.0;
			break;
		case 3:
			ineqvalue = -XX[ii+jj*N] + XX[ii+kk*N] - XX[jj+kk*N] + 1.0;
			break;
		default: // case 4 :
			ineqvalue = -XX[ii+jj*N] - XX[ii+kk*N] + XX[jj+kk*N] + 1.0;
		}

    return ineqvalue;
}

