// *****************************************************************************
// *                                                                           *
// *  BiqCrunch is a semidefinite-based solver for binary quadratic problems.  *
// *  It uses a branch-and-bound method featuring an improved semidefinite     *
// *  bounding procedure, mixed with a polyhedral approach. BiqCrunch uses     *
// *  particular input files format to describe the combinatorial problems.    *
// *                                                                           *
// *   Copyright (C) 2010-2014 Nathan Krislock, Jérôme Malick, Frédéric Roupin *
// *                                                                           *
// *                 http://www-lipn.univ-paris13.fr/BiqCrunch/                *
// *									       *
// *****************************************************************************
//									       *
//    This program is free software: you can redistribute it and/or modify     *
//    it under the terms of the GNU General Public License as published by     *
//    the Free Software Foundation, either version 3 of the License, or        *
//    (at your option) any later version.                                      *
//                                                                             *
//    This program is distributed in the hope that it will be useful,          *
//    but WITHOUT ANY WARRANTY; without even the implied warranty of           *
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
//    GNU General Public License for more details.                             *
//                                                                             *
//    You should have received a copy of the GNU General Public License        *
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.    *
//                                                                             *
// *****************************************************************************

#include <stdio.h>

#include "bb.h"
#include "biqcrunch.h"

extern Problem *SP;
extern FILE *output;


int addSolution(BobNode *node, double sol_value, int * x, int credit) {
	int i;
	int add = 0;

	if (BC_isFeasibleSolution(x)) {
		for (i = 0; i < BobPbSize; i++)
			node->sol.X[i] = x[i];

		if (Bob_ULBUpd((int) sol_value, &(node->sol))) {
// if new solution is better than the global solution,
// then print the new solution
// also, updating the global best solution removes all nodes
// from the queue with lower bounds larger than heur_val

			add = 1;
			switch (credit) {
			case NEWSOL_LEAF:

				fprintf(output, "Leaf node: \t");
				break;
			case NEWSOL_HEUR1:

				fprintf(output, "Heuristic 1: \t");
				break;
			case NEWSOL_HEUR2:

				fprintf(output, "Heuristic 2: \t");
				break;
			case NEWSOL_HEUR3:

				fprintf(output, "Heuristic 3: \t");
				break;

			}
			Bob_PrintSolution();
		}
	}
	return add;
}

/*
 * Print the value of the solution.
 *
 * @param bs: user defined structure that contains all the information 
 *            about the solution
 */
void Bob_PrintSolution() {
	if (SP->max_problem)
		fprintf(output, "Beta updated => %d\n", -Bob_ULBGet());
	else
		fprintf(output, "Beta updated => %d\n", Bob_ULBGet());
}




double BC_evaluateSolution(int *sol) {
	int i;
	double heur_val = 0.0;

	for (i = 0; i < SP->Qs.nnz; i++) {
		int ii = SP->Qs.i[i];
		int jj = SP->Qs.j[i];
		double val = SP->Qs.val[i];

		if (ii < SP->n - 1 && jj < SP->n - 1) {
			heur_val += (double) (2 * sol[ii] - 1.0) * (2 * sol[jj] - 1.0)
					* (val + val);
		} else if (ii == SP->n - 1 && jj < SP->n - 1) {
			heur_val += (double) (2 * sol[jj] - 1.0) * (val + val);
		} else if (ii == SP->n - 1 && jj < SP->n - 1) {
			heur_val += (double) (2 * sol[ii] - 1.0) * (val + val);
		} else if (ii == SP->n - 1 && jj == SP->n - 1) {
			heur_val += val;
		}
	}
	return heur_val;
}


int BC_isFeasibleSolution(int *values) {

	int m, ii, jj, kk;
	int N = SP->n;
    double val;
	double sum;
	double tmp;
	int v[N];
	for (ii = 0; ii < N - 1; ii++)
		v[ii] = 2 * values[ii] - 1;
	v[N - 1] = 1;

	for (m = 0; m < SP->mB; m++) {
		sum = 0.0;

		for (kk = 0; kk < SP->Bs[m].nnz; kk++) {
			ii = SP->Bs[m].i[kk];
			jj = SP->Bs[m].j[kk];
			val = SP->Bs[m].val[kk];

			tmp = (val) * (double) v[ii] * (double) v[jj];
			sum += tmp;
			if (ii != jj)
				sum += tmp;
		}

		if (sum != SP->b[m]) {
			return 0;
		}
	}

	for (m = 0; m < SP->mA; m++) {
		sum = 0;

		for (kk = 0; kk < SP->As[m].nnz; kk++) {
			ii = SP->As[m].i[kk];
			jj = SP->As[m].j[kk];
			val = SP->As[m].val[kk];

			tmp = (val) * (double) v[ii] * (double) v[jj];
			sum += tmp;
			if (ii != jj)
				sum += tmp;
		}

		if (sum > SP->a[m]) {
			return 0;
		}
	}
	return 1;
}

